/* Donated to the public domain 2007,2011 by Sam Trenholme
 *
 * This software is provided 'as is' with no guarantees of correctness or
 * fitness for purpose.
 */

/* This is a tiny "du" for Windows (since MSYS doesn't come with "du", which
 * is one of the handful of commands I miss)
 *
 * sd = "small du"; it is equivalent to "du -ab"
 */

/* Longest path name */
#define PATH_MAX 32768

#include <stdint.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>

/* Fileprint: Print a filename, escaping control characters */
void fileprint(char *filename) {
        printf(" ");
	while(*filename != '\0') {
		if((*filename < 32 && *filename >= 0) 
                   || *filename == '~' || *filename == 127) {
			printf("~%02x",*filename);
		} else {
			printf("%c",*filename);	
		}
		filename++;
	}
	printf("\n");
	fflush(stdout);
}

/* The routine that handles directories */
int64_t dirhandle(char *dir) {
    char path[PATH_MAX];
    DIR *wd;
    struct dirent *f;
    struct _stati64 s;
    int64_t size = 0;

    wd = opendir(dir);
    if(wd == NULL) {
	 int c;
	 printf("0 [[UNABLE TO OPEN DIRECTORY]]");
	 for(c=0;c<36;c++) {printf(" ");}
         fileprint(dir);
         return 0;
         }
    f = readdir(wd);
    while(f != NULL) {
        strncpy(path,dir,PATH_MAX - 10);
        strncat(path,"/",1);
	if((strlen(path) + 10 + strlen(f->d_name)) >= PATH_MAX) {
		return;
	}
	if(strcmp(f->d_name,".") != 0 &&
           strcmp(f->d_name,"..") != 0) {
	    strcat(path,f->d_name);
            _stati64(path,&s);
	    if(S_ISDIR(s.st_mode)) {
		size += dirhandle(path);
	    } else if(S_ISREG(s.st_mode)) {
    		printf("%12lld ",s.st_size);
		fileprint(path);
    		size += s.st_size;
            }
        }
        f = readdir(wd);
        }
    closedir(wd);
    printf("%12lld ",size);
    fileprint(dir);
    return size;
    }

/* The main routine.  This reads files/directories specified on the 
   command line, then makes a hash out of that file. */

main(int argc, char **argv) {
    int counter, q;
    struct _stati64 s;
    /* Check the command line argument */
    if(argc < 2) {
        if(argc >= 1) {
            printf("Usage: %s {dirname}\n",argv[0]);
	    exit(1);
	    }
        else {
	    printf("Usage: <this program> {dirname}\n");
	    exit(2);
	    }
        }
    for(counter = 1; counter < argc; counter++) {	
        _stati64(argv[counter],&s);
	if(S_ISDIR(s.st_mode)) {
		dirhandle(argv[counter]);
	} 
    }
    return 0;
}

